\section{Set representation. Hashing}\label{hash}

\subsection{Hashing with open addressing}

     In section~\ref{types} we considered several representations
for sets whose elements are integers of arbitrary size.
 \index{set!representation}
However, all those representations are rather inefficient:
at least one of the operations (membership test,
adding or deleting an element) runs in time proportional to the
number of elements in the set. This is unacceptable in almost all
practical applications.

It is possible to find a set representation where all three
operations mentioned run in time $C\log n$ (in the worst case).
One such representation is considered in the next section.
In
this section, we consider another set representation that may
require $n$ operations in the worst case but is very efficient
``in a typical case''. The method is
called ``hashing''.
 \index{hashing}

    Suppose we want to store a set of elements of type
\w{T},  where
the number of elements is guaranteed to be less than
\w{n}. Choose a function \w{h} that is defined on elements of
type \w{T} and whose values are integers in the range
$\w{0..n-1}$. It is desirable that this function have
different values for different elements of the set we are trying
to represent (the worst case is when all the function values are
the same). This function is called a {\it hash function}.
 \index{hash function}%

    Our representation uses two arrays
\begin{verbatim}
         val:  array [0..n-1] of T;
         used: array [0..n-1] of Boolean;
\end{verbatim}
(we write \w{n-1} in the type definition, though it is not
permitted in Pascal). The set consists of $\w{val}\,\w{[i]}$
for all \w{i} such that
$\w{used}\,\w{[i]}$
is true. (The values
$\w{val}\,\w{[i]}$ are all
different.) When possible, we
store an element \w{t} at position \w{h(t)}, which
is considered a
``natural place'' for \w{t}. However,
 \index{open addressing}
 \index{hashing!with open addressing}
it may happen that a new element \w{t} appears whose place
\w{h(t)} is already used by another element (that is,
\w{used[h(t)]} is true). In this case, we search to the right looking
for the first unused place and put the element \w{t} there.
(Here ``to the right'' means that the index increases;
when we reach \w{n-1}, the index wraps around.) Recall that we
assume that the number of elements is always less than the number
of places, therefore free places do exist.

     Formally speaking, the invariant relation that we maintain
is the following: For any element, the interval between its
natural place and its actual place is filled completely.

     This invariant makes the membership test easy. Suppose
we want to check if an element \w{t} is in the set. We find the
natural place for \w{t} and then go to the right until we find
an empty slot or \w{t}. In the first case, the element \w{t} is
not in the set (a consequence of our invariant); in the second
case, the element is in the set. If it is absent, we may add it
(filling the unused place found). If not, we can delete it by
putting \w{False} in the corresponding cell of the \w{used} array.

\Problem{%
    The last passage has a severe error. Find it and correct it.}

    \Solution. The delete operation implemented as described can
destroy the invariant and create an empty position between
the
natural and actual positions of some element. We should be more
careful. After a gap appears, we move from left to right until
we find either an element that is not at its natural place or
another gap. If the gap appears first, we have nothing to worry
about. If an element is found not at its natural place, we check
whether it needs to be moved to the gap we've created. If not,
we continue our search. If yes, we move the element found to the
gap. A new gap appears which we deal with in the same way.\es

\Problem{%
    Write the programs for membership test, adding and deleting
elements.}

     \Solution.
\begin{verbatim}
function is_element (t: T): Boolean;
| var i: integer;
begin
| i := h (t);
| while used [i] and (val [i] <> t) do begin
| | i := (i + 1) mod n;
| end; {not used [i] or (val [i] = t)}
| is_element := used [i] and (val [i] = t);
end;
\end{verbatim}

\begin{verbatim}
procedure add (t: T);
| var i: integer;
begin
| i := h (t);
| while used [i] and (val [i] <> t) do begin
| | i := (i + 1) mod n;
| end; {not used [i] or (val [i] = t)}
| if not used [i] then begin
| | used [i] := true;
| | val [i] := t;
| end;
end;

procedure delete (t: T);
| var i, gap: integer;
begin
| i := h (t);
| while used [i] and (val [i] <> t) do begin
| | i := (i + 1) mod n;
| end; {not used [i] or (val [i] = t)}
| if used [i] and (val [i] = t) then begin
| | used [i] := false;
| | gap := i;
| | i := (i + 1) mod n;
| | {gap may be filled by one of i,i+1,...}
| | while used [i] do begin
| | | if i = h (val[i]) then begin
| | | | {i is the natural place, nothing to do}
| | | end else if dist(h(val[i]),i) < dist(gap,i) then begin
| | | | {gap...h(val[i])...i, nothing to do}
| | | end else begin
| | | | used [gap] := true;
| | | | val [gap] := val [i];
| | | | used [i] := false;
| | | | gap := i;
| | | end;
| | | i := (i + 1) mod n;
| | end;
| end;
end;
\end{verbatim}

   Here $\w{dist}\,\w{(a,b)}$ is the distance from
\w{a} to \w{b} measured clockwise, that is,
$$
 \w{dist}\,\w{(a,b)} =  \w{(b}\,\w{-}\,\w{a}\,\w{+}\,\w{n)}\:\w{mod}\:\w{n}.
$$
(We add \w{n}, because \w{mod} works best when the dividend is
positive.)\es

\Problem{%
     There are many versions of hashing. For example, when
we find that the natural place (say, $i$) is occupied, we look
for a free place not among $i+1, i+2,\ldots$, but among $r(i),
r(r(i)), r(r(r(i))),\ldots$ where $r$ is some mapping of the set
$\{0,\ldots,n-1\}$ into itself. What are the possible problems?}

   Answer. (1) We cannot guarantee that free space will be
found even if we know it exists.
(2) It is not clear how to fill gaps after deleting an element.
(In many practical cases, deletion is not necessary, so this approach is
sometimes used. The idea is that a careful choice of the function
$r$ will prevent the appearance of big ``clusters'' of occupied
cells.)\es

\Problem{%
     Suppose hashing is used to store the set of all English
words (say, for a spelling checker). What should we add to the data
to be able to find Russian translations of all English words?}

    \Solution. The array \w{val} (whose elements are English
words) should be extended by a parallel array \w{rval} of their
translations: if \w{used[i]} is true, \w{rval[i]} is a
translation of \w{val[i]}\es

