\subsection{Hashing using lists}

     A hash function with $k$ values is a tool
that reduces the storage problem for one large set to a storage
problem for $k$ small sets. Indeed, after a hash function with $k$
values is chosen, any set is split into $k$ subsets
corresponding to the
$k$ different values of the hash function. (Some of them may be empty.)
If we want to perform a membership test or an add/delete operation,
we compute the hash function value and determine for which of the $k$
sets the operation should be performed.
 \index{set!representation}
 \index{hashing!using lists}

     These smaller sets may be stored conveniently using
references, because we know the total size of all the sets but
not their individual sizes. The following problem suggests
an
implementation.

\Problem{%
     Suppose the values of hash function \w{h} are $\w{1..k}$.
For any number \w{j} in $\w{1..k}$, consider a list of all set
elements \w{z} such that $\w{h(z)}=\w{j}$.
Let us store those \w{k} lists using the variables}
\begin{verbatim}
     Content: array [1..n] of T;
     Next: array [1..n] of 1..n;
     Free: 1..n;
     Top: array [1..k] of 1..n;
\end{verbatim}
in the same way as we did for \w{k} stacks of limited size
(p.~\pageref{pointerlists}).
Write the corresponding procedures. (Please note that deletion is
now easier than in the open addressing case.)

    \Solution. We start with
$\w{Top[i]}=\w{0}$ for all $\w{i}=\w{1..k}$.
All the positions
are linked in a free list as follows:
$\w{Free}=\w{1}$;\quad   $\w{Next[i]}=\w{i+1}$ for
$\w{i}=\w{1}\ldots\w{n-1}$;\quad  $\w{Next[n]}=\w{0}$.
\begin{verbatim}
  function is_element (t: T): Boolean;
  | var i: integer;
  begin
  | i := Top[h(t)];
  | {we should search in the list starting from i}
  | while (i <> 0) and (Content[i] <> t) do begin
  | | i := Next[i];
  | end; {(i=0) or (Content [i] = t)}
  | is_element := (i<>0) and (Content[i]=t);
  end;

  procedure add (t: T);
  | var i: integer;
  begin
  | if not is_element (t) then begin
  | | i := Free;
  | | {Free<>0; we assume that the size limit is not reached}
  | | Free := Next[Free];
  | | Content[i]:=t;
\end{verbatim}
\clearpage
\begin{Verbatim}
  | | Next[i]:=Top[h(t)];
  | | Top[h(t)]:=i;
  | end;
  end;

  procedure delete (t: T);
  | var i, pred: integer;
  begin
  | i := Top[h(t)]; pred := 0;
  | {we should search in the list starting from i;
  |   pred is a predecessor of i in the list
  |   (if exists; otherwise 0)}
  | while (i <> 0) and (Content[i] <> t) do begin
  | | pred := i; i := Next[i];
  | end; {(i=0) or (Content[i] = t)}
  | if i <> 0 then begin
  | | {Content[i]=t, the element exists
  | |   and should be deleted}
  | | if pred = 0 then begin
  | | | {this is the first element in the list}
  | | | Top[h(t)] := Next[i];
  | | end else begin
  | | | Next[pred] := Next[i]
  | | end;
  | | {it remains to return i to the free list}
  | | Next[i] := Free;
  | | Free:=i;
  | end;
  end;
\end{Verbatim}

\Problem{%
    (Requires some probability theory)
Suppose a hash function
with $k$ values is used to store a set of
cardinality $n$. Prove that the expected number of operations in
the preceding problem does not exceed $C(1+n/k)$, if the
element \w{t} is taken at random in such a way that all values
of $h(t)$ are equiprobable (have probability $1/k$).}%
 \index{hashing!running time, upper bound}

    \Solution. Let $l(i)$ be the length of the list corresponding
to the hash value $i$. The number of operations does not exceed
$C(1+l(h(t)))$; the expectation does not exceed $C(1+n/k)$, since
$\sum_i l(i) = n$.\es

\smallskip

   This estimate is based on the assumption that all values of
$h(t)$ have the same
probability. However, for a given input
distribution and a given hash function this assumption may be false,
and many elements of the set may share the same value of the hash
function, so large clusters appear. A method that avoids
this difficulty is called {\it universal hashing}.
 \index{hashing!universal}

The idea is to use a family of hash functions instead of just
one
and to choose a function from this family at random.
The hope is that any fixed set behaves well for most of the
functions in the family.

    Let $H$ be a family of functions. Each function maps the set
$T$ into a set of cardinality $k$ (say, into
 \index{hash function!universal family}%
$0,\ldots,k-1$). The family $H$ is called a {\it universal
family of hash functions\/} if for any two distinct elements
$s,t\in T$, the probability of the event $h(s)=h(t)$ (for a
random function $h\in H$) is equal to $1/k$. (In other words,
the functions $h\in H$ satisfying $h(s)=h(t)$ are in proportion
$1/k$ with all functions in $H$.)

    Remark. A stronger requirement may be given, namely, we may
require that for any two distinct elements $s,t\in H$, the
values $h(s)$ and $h(t)$ (for a randomly chosen $h$) are
independent random variables uniformly distributed among
$0,\ldots,k-1$. This stronger requirement is fulfilled in the
examples below.

\Problem{%
     Let $t_1,\ldots,t_n$ be any sequence of distinct elements
of the set $T$. Consider the sequence of events that occurs when
the elements $t_1,\ldots,t_n$ are added to a set stored using
a hash function $h$ from a universal family $H$. Prove that the
expected number of operations (the mean value taken over all
$h\in H$) does not exceed $Cn(1+n/k)$.}

    \Solution. By $m_i$ we mean the number of elements among
$t_1,\ldots,t_n$ with hash value $i$. (Of course, the numbers
$m_0,\ldots,m_{k-1}$ depend on $h$.) The number of operations we
are interested in is equal to $m_0^2+m_1^2+\cdots+m_{k-1}^2$ up
to a constant factor. (Indeed, if $s$ elements are placed in a
list, the number of operations is approximately
$1+2+\cdots+s\sim s^2$.) The same sum of squares may be written
as the number of pairs $\langle p,q\rangle$ satisfying
$h(t_p)=h(t_q)$. For any fixed $p$ and $q$ the event
$h(t_p)=h(t_q)$ has probability $1/k$ (assuming that $p\ne q$).
Therefore, the expected value of the corresponding term is equal
to $1/k$, and the expected value of the sum is roughly $n^2/k$.
More precisely, we obtain $n + n^2/k$ since we need to count
terms with $p=q$.\es

\smallskip

   This problem shows that the average number of operations per
element is $C(1+n/k)$. Here $n/k$ may be called the
``average load of a hash value''.

\Problem{%
     Prove a similar assertion about the arbitrary sequence
of additions, deletions, and membership tests (not only additions,
as in the preceding problem).}

    \Hint{Let us imagine that while performing
addition, search, or
deletion, the element is a person that
traverses the list of its colleagues
with the same hash value until it finds its twin brother (an equal
element; it this case, the element disappears) or
reaches the end of the list. By $i$-$j$-meeting we mean the
event when elements $t_i$ and $t_j$ meet each other. (It may or may not
happen depending on $h$.) The total number of
operations is (up to a constant factor) equal
to the number of meetings plus the number of elements. When
$t_i\ne t_j$, the probability of an $i$-$j$-meeting does not exceed
$1/k$. It remains to count the meetings of equal elements. Let
us fix some value $x\in T$ and consider all operations that
refer to this value. They follow the pattern: tests, addition,
tests, deletion, tests, addition, etc. The meetings
occur between an added element and tested elements that follow
it (up to the next deletion, and including it), therefore the total
number of meetings does not exceed the number of elements equal
to $x$.}

\smallskip

    Now we give several examples of universal families.
 \index{hash function!universal family}
For any two finite sets $A$ and $B$, the family of all functions
that map $A$ into $B$ is an universal family. However, from a
practical viewpoint this family is useless, since to store a random
function from this family, we need an array with $\#A$ elements
($\#A$ is the cardinality of $A$). If we can afford an array
of that size, we do not need hashing at all!

     More practical examples of universal families may be
obtained using simple algebraic techniques. By
${\Bbb Z}_p$ we denote the set of all residues modulo $p$ where
$p$ is a prime number, that is, the set $\{0,1,\ldots,p-1\}$.
Arithmetic operations are performed on this set modulo $p$.
An universal
family is formed by all linear functionals defined on ${\Bbb
Z}_p^n$ with values in ${\Bbb Z}_p$. More precisely, let
$a_1,\ldots,a_n$ be arbitrary elements of ${\Bbb Z}_p$ and consider
the mapping
 %
$$
   h: \langle x_1,\ldots,x_n \rangle \mapsto
     a_1x_1+\cdots+a_nx_n
$$
We get a family of $p^n$ mappings ${\Bbb Z}^n_p \to {\Bbb Z}_p$
indexed by $n$-tuples $\langle a_1,\ldots,a_n
\rangle$.

\Problem{%
     Prove that this family is universal.}

    \Hint{Let $x$ and $y$ be distinct points of the space
${\Bbb Z}_p^n$. What is the probability of the event ``a random
functional $\alpha$ has the same values for $x$ and $y$''? In
other words, what is the probability of the event
``$\alpha(x-y)=0$''? The answer is provided by the following
statement. If $u$ is a nonzero vector, all possible
values of $\alpha(u)$ are equiprobable.}

\smallskip

    In the following problem, the set ${\Bbb B}=\{0,1\}$
is taken to be the set of residues modulo $2$.

\Problem{%
    Show that the family of all linear mappings of ${\Bbb B}^n$ into
${\Bbb B}^m$ is universal.\nobreak\es}

\medskip

     Hashing turns out to be useful in unexpected
curcumstances.
The following example was communicated to me
by D.V.~Varsonofiev.
 \index{Varsonofiev, D.V.}
Suppose we want to contruct a spelling checker to find
(most of) the typos in an English text. We do not want, however,
 \index{spelling checker}
to keep a list of all correct words (in all grammatic
forms). We can use the
following trick. Choose some positive integer
$N$ and functions
$f_1,\ldots,f_k$ that map words to
$1,\ldots,N$. Consider an array of $N$ bits initially
set to
zero. Then
for any (correctly spelled) word $x$, compute the values
$f_1(x),\ldots,f_k(x)$ and make the corresponding bits equal to $1$.
(Some bits may correspond to several words.) Then the approximate test
to check whether a string $z$ is a correctly spelled word, is as
follows. Compute all values
$f_1(z),\ldots,f_k(z)$ and check that all the corresponding bits are
$1$s. This test may miss some errors, but all correct
words will be allowed.

