\subsubsection*{Tree representation}

     The simplest way to represent a tree
is to identify the vertices of a full
binary tree with integers $1,2,3,\ldots$ (the left son of $n$ is
$2n$, the right son of $n$ is $2n + 1$) and store the labels in
an array \verb|val [1..N]| (for a large enough \w{N}).
However, this approach wastes space
because space is set aside for positions in the full binary
tree that are not filled in a specific
$T$-tree.

     The following approach is more space efficient.
We use three arrays
 \index{tree!pointer implementation}
\begin{verbatim}
       val: array [1..n] of T;
       left, right: array [1..n] of 0..n;
\end{verbatim}
(\verb|n| is the maximal possible number of tree vertices for
trees we want to store) and a variable \verb|root:0..n|. Each
vertex of the stored $T$-tree will have a number that is an
integer in \w{1..n}. Different vertices have different
numbers; some numbers may be unused. The label of the vertex
with number \w{x} is stored in \verb|val[x]|. The root has
number \verb|root|. If vertex \verb|i| has sons, their
numbers are \verb|left[i]| and \verb|right[i]|. Nonexistent
sons are replaced by the number \verb|0|. Similarly, the condition
$\w{root}=\w{0}$ means that the tree is empty.

     The tree vertices only occupy part of the array. For ``free''
values of \w{i} that are not used as vertex numbers,
the values \verb|val[i]| have no meaning. We want the free
numbers to be ``linked in a list''; the first free number is
stored in a special variable \verb|free:0..n|, while the free
number that follows \verb|i| in the list is \verb|left[i]|. In
other words, the list of all free numbers is
  %
\begin{center}
   \verb|free, left[free], left[left[free]],...|
\end{center}
  %
For the last free number \verb|i| in the list, the value
\w{left[i]} equals \w{0}. If $\w{free}=\w{0}$,
there are no free numbers. (Remark. We used the array \w{left} to
link all free numbers in a list but of course, we may use the
array \w{right} instead.)

     We can use any other integer outside \w{1..n}
to indicate the absence of a vertex
(instead of \w{0}).
To stress this, we use a symbolic
constant $\w{null}=\w{0}$ instead of the numeral \w{0}.

     \Problem{Write a procedure that checks if an element \w{t:T} is
present in an ordered tree (as described above).}

     \Solution.
\begin{Verbatim}
  if root = null then begin
  | ..t is not in the tree
  end else begin
  | x := root;
  | {invariant: it remains to check if t is present in
  |   a nonempty subtree rooted at x}
  | while ((t < val [x]) and (left [x] <> null)) or
  | |     ((t > val [x]) and (right [x] <> null)) do begin
  | | if t < val [x] then begin {left [x] <> null}
  | | | x := left [x];
  | | end else begin {t > val [x], right [x] <> null}
  | | | x := right [x];
  | | end;
  | end;
  | {either t = val [x] or t is not in the tree}
  | ..answer is (t = val [x])
  end;
\end{Verbatim}

     \Problem{Simplify the procedure using the following trick.
Extend the array \w{val}, adding a cell with index \w{null}.
Let \w{val[null]} be \w{t}.}

     \Solution.
\begin{Verbatim}
  val [null] := t;
  x := root;
  while t <> val [x] do begin
  | if t < val [x] then begin
  | | x := left [x];
  | end else begin
  | | x := right [x];
  | end;
  end;
  ..answer is (x <> null).
\end{Verbatim}

     \Problem{Write a procedure that adds an element \verb|t|
to a
set represented as an (ordered) $T$-tree. (If \verb|t| is already
present, nothing should be done.)}

     \Solution. The procedure \verb|get_free (var i:integer)|
produces a free integer \verb|i| in \w{1..n} (that is, an integer
that is not a number of
any vertex) and updates the free list. (For simplicity,
we assume that free integers
exist.)
\begin{verbatim}
  procedure get_free (var i: integer);
  begin
  | {free <> null}
  | i := free;
  | free := left [free];
  end;
\end{verbatim}
Using this procedure, we write:
\begin{verbatim}
  if root = null then begin
  | get_free (root);
  | left [root] := null; right [root] := null;
  | val [root] := t;
  end else begin
  | x := root;
  | {invariant: it remains to add t to a (nonempty) subtree
  |  rooted at x}
  | while ((t < val [x]) and (left [x] <> null)) or
  | |     ((t > val [x]) and (right [x] <> null)) do begin
  | | if t < val [x] then begin
  | | | x := left [x];
  | | end else begin {t > val [x]}
  | | | x := right [x];
  | | end;
  | end;
\end{verbatim}
\clearpage
\begin{Verbatim}
  | if t <> val [x] then begin {t is not in the tree}
  | | get_free (i);
  | | left [i] := null; right [i] := null;
  | | val [i] := t;
  | | if t < val [x] then begin
  | | | left [x] := i;
  | | end else begin {t > val [x]}
  | | | right [x] := i;
  | | end;
  | end;
  end;
\end{Verbatim}

    \Problem{Write a procedure that deletes an element \w{t} from a
set represented as an ordered tree. (If the element is not in the set,
nothing should be done.)}

     \Solution.
\begin{verbatim}
if root = null then begin
| {the tree is empty, there is nothing to do}
end else begin
| x := root;
| {it remains to delete t from the subtree rooted at x;
|  since it may require changes in the father node,
|  we introduce the variables  father: 1..n and
|  direction: (l, r) with the following
|  invariant: if x is not the root, then father
|  is (the number of) x's father node, direction is
|  equal to l/r if x is the left/right son of its father}
| while ((t < val [x]) and (left [x] <> null)) or
| |     ((t > val [x]) and (right [x] <> null)) do begin
| | if t < val [x] then begin
| | | father := x; direction := l;
| | | x := left [x];
| | end else begin {t > val [x]}
| | | father := x; direction := r;
| | | x := right [x];
| | end;
| end;
| {t = val [x] or t is not in the tree}
| if t = val [x] then begin
| | ..delete the node x with a known father and direction
| end;
end;
\end{verbatim}
The deletion of a vertex uses the procedure
\begin{verbatim}
  procedure make_free (i: integer);
  begin
  | left [i] := free;
  | free := i;
  end;
\end{verbatim}
which adds the number \verb|i| to the free list.
While deleting a vertex, we should distinguish between
four cases depending on whether the vertex has a
left/right son or not.
\begin{verbatim}
if (left [x] = null) and (right [x] = null) then begin
| {x is a leaf, no sons}
| make_free (x);
| if x = root then begin
| | root := null;
| end else if direction = l then begin
| | left [father] := null;
| end else begin {direction = r}
| | right [father] := null;
| end;
end else if (left[x]=null) and (right[x] <> null) then begin
| {when x is deleted, right[x] occupies its place}
| make_free (x);
| if x = root then begin
| | root := right [x];
| end else if direction = l then begin
| | left [father] := right [x];
| end else begin {direction = r}
| | right [father] := right [x];
| end;
end else if (left[x] <> null) and (right[x]=null) then begin
| ..the symmetrical code
end else begin {left [x] <> null, right [x] <> null}
| ..delete a vertex with two sons
end;
\end{verbatim}
The deletion of a vertex with two sons is the most difficult case.
Here we should exchange it with an immediately
following vertex
(in the sense of label ordering).
\begin{verbatim}
  y := right [x]; father := x; direction := r;
  {now father and direction refer to vertex y}
  while left [y] <> null do begin
  | father := y; direction := l;
  | y := left [y];
  end;
\end{verbatim}
\clearpage
\begin{Verbatim}
  {val[y] is minimal element of the set larger
   than val[x], y has no left son}
  val [x] := val [y];
  ..delete the vertex y (we already know how to do
    that for a vertex without the left son)
\end{Verbatim}

    \Problem{Simplify the deletion procedure using the following
observation: Some cases (say, the first two) may be combined into
a single case.\es}

    \Problem{Use an ordered tree to store a function whose domain is
a finite subset of $T$ and whose range is some set $U$. The
operations are: find the value of the function
for a given argument; change this
value; delete an element from the domain; and add an element to
the domain (the value is also provided).}

     \Solution. We represent the domain using an ordered tree and
add one more array
\begin{center}
   \verb|func_val: array [1..n] of U;|
\end{center}
If \verb|val[x] = t| and \verb|func_val[x] = u|, then the function
value on \verb|t| equals~\w{u}.\es

    \Problem{Assume that we want to find the $k$-th element
of a set (according to the ordering on $T$) in time limited by
$C\cdot(\hbox{\rm tree height})$. What additional information do
we need to store at the tree vertices?}%
 \index{search!$k$-th element}

    \Solution. At each vertex, we store the number of its
descendants. When a vertex is added or deleted, this information
must be updated along a path from the root to the
new/deleted vertex. While searching for the $k$-th vertex,
we maintain the following invariant: the vertex in question is
the $s$-th vertex (according to the  $T$-ordering) of a subtree
rooted at $x$ (here $s$ and $x$ are variables).\es


\subsubsection*{Running time}

    All of the procedures discussed above (membership test,
addition, and deletion) run in time $C\cdot(\hbox{\rm tree
height})$. For a ``well-balanced'' tree where all leaves have
approximately
the same height, the tree height is close to the logarithm of
the number of vertices. However, for a unbalanced tree the
height may be much larger. In the worst case, the vertices may
form a chain (if all vertices have no left son, for example) and
the tree height is the number of vertices. This happens if we
start with the empty set and add elements in increasing
order. However, one can prove that if the elements are added in
random order, then the expected height of the tree will not
exceed $C\log(\hbox{\rm tree size})$. If this ``average
bound'' is not good
enough for our application, we must spend additional
effort to keep the tree ``balanced''. This is explained in the
next section.

