\section{Data types}\label{types}

\subsection{Stacks}

     Let \w{T} be some type. Consider the data type ``stack of elements
of type \w{T}''. Values of that
 \index{stack}%
type are sequences of values of type~\w{T}.

\medskip

     Operations:

\begin{itemize}
\item Make\_empty (\w{var s}: stack of elements of type  \w{T})
\item Add (\w{t:T}; \w{var s}: stack of elements of type \w{T})
\item Take (\w{var t:T}; \w{var s}: stack of elements of type \w{T})
\item Is\_empty (\w{s}: stack of elements of type \w{T}): \w{Boolean}
\item Top (\w{s}: stack of elements of type \w{T}): \w{T}
\end{itemize}

\noindent
(We use Pascal notation even though the stack type
does not exist in Pascal.) The procedure ``Make\_empty'' makes
the stack empty. The procedure ``Add'' adds \w{t} to the end of
the sequence \w{s} (i.e., the top of the stack).
The procedure ``Take'' is applicable if the
sequence \w{s} is not empty; it takes the last element away
from \w{s}
and
puts it into the variable \w{t}. The expression
``Is\_empty(\w{s})'' is true when the sequence \w{s} is empty.
The expression ``Top(\w{s})'' is defined when \w{s} is not
empty; its value is the last element of the sequence \w{s}.

Usually the operations ``Add'' and ``Take'' are called
``Push'' and ``Pop'' respectively; we use the names ``Add''
and ``Take'' to stress the similarity between stacks and
queues (section~\ref{queue}).

     Our goal is to show how stacks can be implemented in Pascal and
what they can be used for.

\subsubsection*{Stack: array implementation}

     Assume that the number of elements in a stack never
exceeds some constant \w{n}. Then the stack can be implemented
 \index{stack!array implementation}%
using two variables:
\begin{verbatim}
        Content: array [1..n] of T;
        Length: integer;
\end{verbatim}
We assume that our stack contains elements
\begin{verbatim}
        Content [1],...,Content [Length]
\end{verbatim}

\begin{itemize}
\item
To make the stack empty, it is enough to perform the assignment
\begin{verbatim}
        Length := 0
\end{verbatim}
\item
     Adding element \w{t}:
\begin{verbatim}
        {Length < n}
        Length := Length + 1;
        Content [Length] :=t;
\end{verbatim}
\item
     Taking element into a variable \w{t}:
\begin{verbatim}
         {Length > 0}
         t := Content [Length];
         Length := Length - 1;
\end{verbatim}
\item
     The stack is empty when \w{Length = 0}.
\item
     The top of the stack is \w{Content [Length]}, assuming \w{Length > 0}.
\end{itemize}

Therefore, a variable of type stack can be replaced in
a
Pascal program
by two variables \w{Content} and \w{Length}.
We can also define the type  \w{stack} as follows:
\begin{verbatim}
    const n = ...
    type
      stack = record
      | Content: array [1..n] of T;
      | Length: integer;
      end;
\end{verbatim}
We then define procedures dealing with stack
variables. For example, we write
\begin{verbatim}
      procedure Add (t: T; var s: stack);
      begin
      | {s.Length < n}
      | s.Length := s.Length + 1;
      | s.Content [s.Length] := t;
      end;
\end{verbatim}

\subsubsection*{The use of stacks}

     In the following problem, we consider sequences of opening
and closing parentheses \w{( )} and square brackets \w{[ ]}. Some
sequences are considered to be ``correct''.
 \index{parentheses!correct expressions}%
Namely, a sequence is correct if its correctness follows from
the following rules:

\begin{itemize}
   \item the empty sequence is correct;
   \item if $A$ and $B$ are correct, then $AB$ is correct;
   \item if $A$ is correct, then \w{[}$A$\w{]} and \w{(}$A$\w{)}
         are correct.
\end{itemize}

     Example. The sequences \w{()}, $\w{[[}\,\w{]]}$,
$\w{[()[}\,\w{]()][}\,\w{]}$ are correct, while the sequences
\w{]}, \w{)(}, \w{(]}, \w{([)]} are not.

\Problem{%
    Check the correctness of a given sequence. The number of operations
should be proportional to the length of the sequence. We assume that the
sequence terms are encoded as follows:}
$$\begin{tabular}{rr}
     \w{(} &  $1$ \\
     \w{[} &  $2$ \\
     \w{)} & $-1$ \\
     \w{]} & $-2$
\end{tabular}$$

     \Solution. Let \w{a[1]..a[n]} be a sequence of length \w{n}.
Consider a stack whose elements are opening parentheses and
brackets (i.e.,
the numbers
$1$ and $2$).

     Initially the stack is empty. We scan the sequence from
left to right. When an opening parenthesis or bracket is found,
we put it onto the stack. When a closing parenthesis or bracket
is found, we check if the top of the stack is a complementary
parenthesis or bracket. If not, we stop and reject the input.
If so, we take the top of the stack away. The
sequence is correct if it is not rejected while reading the input
and if
the stack is empty after the input is exhausted.
\begin{verbatim}
        Make_empty (s);
        i := 0; Error_found := false;
        {i symbols are processed}
        while (i < n) and not Error_found do begin
        | i := i + 1;
        | if (a[i] = 1) or (a[i] = 2) then begin
        | | Add (a[i], s);
        | end else begin  {a[i] is either -1 or -2}
        | | if Is_empty (s) then begin
        | | | Error_found := true;
        | | end else begin
        | | | Take (t, s);
        | | | Error_found := (t <> - a[i]);
        | | end;
        | end;
        end;
        Correct := (not Error_found) and Is_empty (s);
\end{verbatim}

       Let us prove the correctness of our program.

(1) If the input sequence is correct, our program accepts it. This
can be proved by induction.
We need to prove that
   %
{\bf (a)}
our
program accepts the empty
sequence;
   %
{\bf (b)} that it accepts the sequence $AB$ (assuming that $A$
and $B$ are accepted);
    %
and
{\bf (c)}
it accepts
the sequences \w{[}$A$\w{]} and \w{(}$A$\w{)} assuming that $A$
is accepted.

An empty sequence is accepted for obvious reasons.
(Note: In this case, the \w{while}-loop is not
executed.)

For $AB$ our program works exactly as for
$A$ until all symbols of
$A$ are processed; therefore,
the stack is empty at that moment. Then
program processes $B$ (and finishes with the empty stack,
because $B$ is accepted by assumption).

For \w{[}$A$\w{]} the program begins by putting
an opening bracket onto the stack. Then the program processes
$A$, the only difference is that there is an additional
bracket
at the bottom of the stack. When $A$ is finished, the
stack is empty except for the left bracket; at the next
step, the stack becomes empty. A similar thing
happens for \w{(}$A$\w{)}.

     (2) Let us now prove that if the program accepts some
sequence, then the sequence is correct. This is proved by
induction over the length of the sequence. Consider the
length of the stack during execution. If
the stack
becomes empty at
some
point,
then the sequence can be divided into two parts and
each of the parts is accepted by the program. Therefore, each
part is correct (inductive hypothesis) and the sequence as a
whole is correct (definition of correctness). Now assume
that the stack never becomes empty (except for the beginning and
the end). This means that the bracket or parenthesis put
onto the stack at the first step is removed at the last
step. Therefore, the first and last symbols in our sequence are
complementary, the sequence is of type \w{(}$A$\w{)} or
\w{[}$A$\w{]}, and the behavior of the program differs from its
behavior on $A$ only by the additional parenthesis or bracket
at the bottom of the stack. Therefore, by the
induction hypothesis,
$A$ is correct and the sequence is correct by
definition.\es

\Problem{%
     The program can be simplified if the sequence contains only
parentheses and no brackets. How?
}

     \Solution. In this case, the stack is reduced to its length,
and we arrive at the following statement: A sequence of ``\w{(}'' and
``\w{)}'' is correct if and only if each prefix contains no more
symbols ``\w{)}'' than ``\w{(}'', and the entire sequence has equal
numbers of both symbols.\es

\Problem{%
     Implement two stacks using one array. The total number of
elements in both stacks is limited by the array length; all stack
operations should run in $O(1)$ time (i.e., running time should be
bounded by a constant).}%
 \index{stack!two in an array}%

     \Solution. The stacks grow from the beginning and end
of the array \w{Content[1..n]} in opposite directions. One stack
occupies places
$$
      \w{Content[1]..Content[Length1]},
$$
while the other stack occupies places
$$
     \w{Content[n]..Content[n-Length2+1]}
$$
(both stacks are listed from bottom to top). Stacks do not
overlap if their total length does not exceed \w{n}.\es

\Problem{%
     Implement \w{k} stacks of elements of type \w{T} with a
total of at most \w{n} elements using arrays with total length $C
(\w{n}+\w{k})$. Each stack operation (except initialization,
which makes all stacks empty) should be performed in
constant time (not depending on \w{n} and \w{k}).
(In other words,
the implementation should
require
space $O(\w{n}+\w{k})$ and
run in
time $O(1)$ for each operation.)}%
 \index{stack!pointer implementation}%
 \label{pointerlists}

     \Solution. We use a ``pointer implementation'' of stacks. It
uses three arrays:
\begin{verbatim}
      Content: array [1..n] of T;
      Next: array [1..n] of 0..n;
      Top: array [1..k] of 0..n;
\end{verbatim}

     The array \w{Content} can be thought of as \w{n} cells
numbered from \w{1} to \w{n}. Each of the cells is capable of holding
one element of type \w{T}. The array \w{Next}
is represented
by arrows between elements:
there is an arrow from \w{i} to \w{j} if
\w{Next[i]=j}. (If \w{Next[i]=0}, there are no arrows
from \w{i}.)
The content of the \w{s}-th stack
($\w{s}\in \{\w{1..k}\}$)
is determined
as follows: the top element is \w{Content[Top[s]]} and
other elements are read by
following the arrow links
(if they exist). Moreover,
$$
\hbox{\rm (\w{s}-th stack is empty)}\;\Leftrightarrow\;\w{Top[s] = 0}.
$$

{\samepage
     The ``arrow trajectories'' starting from
$$
 \w{Top[1]},\ldots, \w{Top[k]}
$$
(those not equal to $0$) are disjoint. Besides these, we
need one more trajectory that traverses all locations that
are currently free. Its
starting point
is stored in the variable
\verb|Free: 0..n| (where $\w{Free}=\w{0}$ means that all places are
occupied). Here is an example:\par
\vspace*{-1ex}
$$
  \setlength{\unitlength}{.8cm}
  \begin{picture}(14, 7)
    %  \put(0,0){\framebox(14,7){}}
       \put(1,1){\makebox(3,1){\tt Top}}
       \put(1,4){\makebox(3,1){\tt Content}}
       \put(9,1){\makebox(3,1){\tt Free}}

       \put(5,1){\framebox(1,1){}}
       \put(6,1){\framebox(1,1){}}

       \put(5,4){\framebox(1,1){  \strut\tt  a}}
       \put(6,4){\framebox(1,1){  \strut\tt  p}}
       \put(7,4){\framebox(1,1){  \strut\tt  q}}
       \put(8,4){\framebox(1,1){  \strut\tt  d}}
       \put(9,4){\framebox(1,1){  \strut\tt  s}}
       \put(10,4){\framebox(1,1){ \strut\tt  t}}
       \put(11,4){\framebox(1,1){ \strut\tt  v}}
       \put(12,4){\framebox(1,1){ \strut\tt  w}}

       \put(5.5,2){\bezier{100}(0,0)(0,1)(0,2)}
       \put(6.5,2){\bezier{300}(0,0)(2.5,1)(5,2)}
       \put(11,2){\bezier{150}(0,0)(0.75,1)(1.5,2)}


       \put(5.5,5){\bezier{120}(0,0)(2,1)(2,0)}
       \put(7.5,5){\bezier{180}(0,0)(3,1.5)(3,0)}
       \put(10.5,5){\bezier{240}(0,0)(-4,2)(-4,0)}
       \put(12.5,5){\bezier{240}(0,0)(-4,2)(-4,0)}
       \put(11.5,5){\bezier{120}(0,0)(-2,1)(-2,0)}


    \end{picture}
$$
$$
\begin{tabular}{r|cccccccc}
\w{Content} &\w{a}&\w{p}&\w{q}&\w{d}&\w{s}&\w{t}&\w{v}&\w{w}\\
\w{Next}  &\w{3}&\w{0}&\w{6}&\w{0}&\w{0}&\w{2}&\w{5}&\w{4}\\
\w{Top}  &\w{1}&\w{7}
\end{tabular}
$$
\vspace*{-.8ex}
$$
  \w{Free} = \w{8}
$$
}
Stacks: the first one contains \w{p}, \w{t}, \w{q}, \w{a} (\w{a} is on the
top); the second one contains \w{s}, \w{v} (\w{v} is on the top).

\begin{verbatim}
  procedure Initialize; {Make all stacks empty}
  | var i: integer;
  begin
  | for i := 1 to k do begin
  | | Top [i]:=0;
  | end;
  | for i := 1 to n-1 do begin
  | | Next [i] := i+1;
  | end;
  | Next [n] := 0;
  | Free:=1;
  end;
\end{verbatim}

\begin{verbatim}
  function Is_free: Boolean;
  begin
  | Is_free := (Free <> 0);
  end;
\end{verbatim}

\begin{verbatim}
  procedure Add (t: T; s: integer);
  | {Add t to the s-th stack}
  | var i: 1..n;
  begin
  | {Is_free}
  | i := Free;
  | Free := Next [i];
  | Next [i] := Top [s];
  | Top [s] :=i;
  | Content [i] := t;
  end;
\end{verbatim}

\begin{verbatim}
  function Is_empty (s: integer): Boolean;
  | {s-th stack is empty}
  begin
  | Is_empty := (Top [s] = 0);
  end;
\end{verbatim}

\begin{verbatim}
  procedure Take (var t: T; s: integer);
  | {Take the top of the s-th stack into t}
  | var i: 1..n;
  | begin
  | {not Is_empty (s)}
  | i := Top [s];
  | t := Content [i];
  | Top [s] := Next [i];
  | Next [i] := Free;
  | Free := i;
  end;
\end{verbatim}

\begin{Verbatim}
  function Top_element (s: integer): T
  | {Top of the s-th stack}
  begin
  | Top_element := Content[Top[s]];
  end;
\end{Verbatim}

