\section{Recursive and nonrecursive programs}\label{delrec}

     For a universal programming language (like Pascal)
recursion is, in a sense, redundant: for any recursive program
it is possible to write an equivalent program without
recursion.%
 \index{recursion!elimination}
Of course, this does not mean that recursion should be avoided,
because it allows us to provide elegant solutions to
otherwise complicated problems.

    However, we want to show some methods that allow us to
eliminate recursion in some cases and transform a recursive
program into an equvalent nonrecursive program.

     What for? A pragmatic answer is that many
computers do not like recursion: recursive programs may be
two or three times slower than equivalent nonrecursive
programs. (Unfortunately, this is the case for some modern
computers using so-called RISC processors.) Another problem is
that some programming languages do not allow
recursion at all. But the main reason is that elimination of
recursion is sometimes
very instructive.

\subsection{Table of values (dynamic programming)}

\Problem{%
     The following recursive procedure computes binomial
coefficients. Write an equivalent program without recursion.}%
 \index{programming!dynamic}%
 \index{dynamic programming}
\begin{verbatim}
        function C(n,k: integer):integer;
        | {n >= 0; 0 <=  k <=n}
        begin
        | if (k = 0) or (k = n) then begin
        | | C:=1;
        | end else begin {0<k<n}
        | | C:= C(n-1,k-1)+C(n-1,k)
        | end;
        end;
\end{verbatim}

Remark. $C(n,k) = {n \choose k}$ is the number of $k$-element subsets
of an $n$-element set.
 \index{binomial coefficient}%
The identity
 $$
   {n \choose k} = {{n-1}\choose{k-1}} + {{n-1}\choose k}
 $$
is proved as follows: Fix some element $x$ of the $n$-element set. Then
all $k$-element subsets are divided into two categories: those that
contain $x$ and those that do not. The elements of the first
type are in one-to-one correspondence with the $(k-1)$-element subsets of
a $(n-1)$-element set (just discard $x$); the elements of the second
type are $k$-element subsets of a $(n-1)$-element set.

The table of ${\displaystyle {n \choose k}}$
$$
\begin{tabular}{ccccccccccc}

               & & & & &1& & & & & \\
               & & & &1& &1& & & & \\
               & & &1& &2& &1& & & \\
               & &1& &3& &3& &1& & \\
               &1& &4& &6& &4& &1& \\
              .&.&.&.&.&.&.&.&.&.&.
\end{tabular}
$$
is called the {\it Pascal triangle} (the same Blaise Pascal who
gave his name to a programming language).%
 \index{Pascal triangle}%
 \index{triangle, Pascal}%
 \index{Pascal, B.}
In this triangle, any element (except the $1$s on the left and
the right) is the sum of the two elements above it.

    \Solution. One may use the formula
$$
   {n\choose k}  = \frac{n!}{k!\,(n-k)!}
$$
We do not use it because we want to show more general methods
to eliminate recursion. Our program fills the table of values
$\w{C(n,k)}= {n\choose k}$ for $n = 0, 1, 2,\ldots$ until it
reaches the element in question.\es

\Problem{%
     Compare the computation time for the recursive and nonrecursive
versions of the binomial coefficient algorithm, and similarly
for the amount of
memory used.}

     \Solution. The table used in the nonrecursive version
occupies space of order $n^2$. We can reduce
it to $n$ if we recall that only one line of the Pascal triangle is
needed to compute the next line. The time required is still $n^2$.

The recursive program requires much more time. Indeed, the call
\w{C(n,k)} causes two calls of type
\w{C(n-1,..)}, those two calls cause four calls of type
\w{C(n-2,..)}, etc. Hence, the time is exponential (of order
$2^n$). The recursive procedure  uses $O(n)$ memory (we have to
multiply the recursion depth, that is $n$,
by the amount of memory required
by one copy of the procedure, that is  $O(1)$.\es

The reason why the nonrecursive version is so much faster
is the following.
In the recursive version, the same
computations are repeated many times. For example, the call
\w{C(5,3)} causes {\it two\/} calls of \w{C(3,2)}:
$$
\def\n#1{\hbox  to 0pt{\hss\w{#1}\hss}}
\begin{array}{ccccccccc}
          &  &       &  &\n{C(5,3)}&  &      &  &                     \\
          &  &       &\swarrow &      &\searrow&      &  &            \\
          &  &\n{C(4,2)}&  &      &  &\n{C(4,3)}&  &                  \\
          &\swarrow&       &\searrow&      &\swarrow&      &\searrow& \\
\n{C(3,1)}&  &    & &\n{C(3,2)}&  &      &  &\n{C(3,3)}
\end{array}
$$

When we fill the table, we compute the value for each cell only
once, hence the economy of the nonrecursive method.
This method is called {\it dynamic
programming}, and is useful when the amount of information to be stored
in the table is not too large.%
 \index{programming!dynamic}%
 \index{dynamic programming}%

\Problem{%
     Compare the recursive and  the (simplest) nonrecursive algorithm
to compute the Fibonacci numbers defined as the sequence}%
 \index{Fibonacci numbers}
$$
  \Phi_1=\Phi_2=1;\quad\Phi_n=\Phi_{n-1}+\Phi_{n-2}\ (n>2). \eqno\es
$$

\Problem{%
     A convex polygon with $n$ vertices is given (by a list of coordinates
of its vertices). It is cut into triangles by non-intersecting
diagonals; to do this, we need exactly $n-3$ diagonals (proof by
induction over $n$). The cost of the
triangulation is defined as the total
length of all the
diagonals used. Find the minimal cost of the triangulation.
The number of operations should be limited by a polynomial of $n$.
(This requirement excludes exhaustive search, since the number
of possibilities is not bounded by any polynomial.)}%
 \label{triangulation}%
 \index{Catalan numbers}%
 \index{triangulation of a polygon}

     \Solution. Assume that the vertices are numbered
from $1$ to
$n$ and the numbers increase in the clockwise direction. Let $k$ and
$l$ be two numbered vertices and suppose $l>k$. By $A(k,l)$ we
denote a polygon cut from the original
polygon by segment
$k$--$l$. (The segment $k$--$l$ cuts the polygon into two
polygons, one of which contains the $1$--$n$ side; $A(k,l)$ is
the other one.) The initial polygon is denoted by $A(1,n)$. When
$l=k+1$,
we have a degenerate polygon with only two vertices.
 $$
\setlength{\unitlength}{6mm}
\begin{picture}(10,7)
  \put(3,1){\circle*{0.1}}
   \put(3,1){\line(1,0){2}}
  \put(5,1){\circle*{0.1}}
   \put(5,1){\line(2,1){2}}
  \put(7,2){\circle*{0.1}}
   \put(7,2){\line(1,1){1}}
  \put(8,3){\circle*{0.1}}
   \put(8,3){\line(0,1){1}}
  \put(8,4){\circle*{0.1}}
   \put(8,4){\line(-1,2){1}}
  \put(7,6){\circle*{0.1}}
   \put(7,6){\line(-1,0){2}}
  \put(5,6){\circle*{0.1}}
   \put(5,6){\line(-2,-1){2}}
  \put(3,5){\circle*{0.1}}
   \put(3,5){\line(-1,-2){1}}
  \put(2,3){\circle*{0.1}}
   \put(2,3){\line(1,-2){1}}

   \put(2,3){\line(1,0){6}}

   \put(2,0){\makebox(1,1){$1$}}
   \put(1,2.5){\makebox(1,1){$k$}}
   \put(8,2.5){\makebox(1,1){$l$}}
   \put(5,0){\makebox(1,1){$n$}}

   \put(5,4){\makebox(1,1){$A(k,l)$}}

\end{picture}
 $$
By $a(k,l)$ we denote the minimal cost of triangulation of
$A(k,l)$. Let us write a recurrence formula for $a(k,l)$. When
$l=k+1$, we have a degenerate polygon with two vertices and let
$a(k,l)=0$. When $l=k+2$, we have a triangle, and $a(k,l)=0$.
Assume that $l>k+2$.
 $$
\setlength{\unitlength}{6mm}
\begin{picture}(10,5)
  \put(8,1){\circle*{0.1}}
   \put(8,1){\line(0,1){1}}
  \put(8,2){\circle*{0.1}}
   \put(8,2){\line(-1,2){1}}
  \put(7,4){\circle*{0.1}}
   \put(7,4){\line(-1,0){2}}
  \put(5,4){\circle*{0.1}}
   \put(5,4){\line(-2,-1){2}}
  \put(3,3){\circle*{0.1}}
   \put(3,3){\line(-1,-2){1}}
  \put(2,1){\circle*{0.1}}

   \put(2,1){\line(1,0){6}}

   \put(1,0.5){\makebox(1,1){$k$}}
   \put(8,0.5){\makebox(1,1){$l$}}

   \put(4.5,4){\makebox(1,1){$i$}}
   \put(2,1){\line(1,1){3}}
   \put(8,1){\line(-1,1){3}}

\end{picture}
$$
The chord $k$--$l$ is a side of the polygon $A(k,l)$; therefore,
it is a side of one of the triangles of the triangulation. The
opposite vertex of this triangle has some number $i$. It may be
any of the vertices $k+1,\ldots,l-1$, and the minimal
triangulation cost can be computed as
 $$
   \min\{\hbox{(the length of $k$--$i$)}
       +\hbox{(the length of $i$--$l$)}+a(k,i)+a(i,l)\}
 $$
where the minimal value is taken over all $i=k+1,\ldots,l-1$. We
should also
take into account that for $q=p+1$, the segment
$p$--$q$ is one of the sides and its length should be
counted as
$0$ for
our purposes.

   This formula allows us to fill the table of values $a(k,l)$
in order of increasing number of vertices (which is $l-k+1$).
The corresponding program uses memory of order $n^2$ and time of
order $n^3$ (one application of the recurrent formula requires
a search for a minimal value
among not more than $n$ numbers).\es

\Problem{%
      An $m\times n$ matrix is a rectangular table with $m$
rows and $n$ columns filled with numbers. An $m\times n$ matrix may
be multiplied by an $n\times k$ matrix (the width of the left
factor must be equal to the height of the matrix on the right)
giving
$m\times k$ matrix as the result. The cost of such a
multiplication is defined as $mnk$ (this is the number of
multiplications required by the standard
multiplication algorithm, but this is not important).
Matrix multiplication is associative, therefore the
product of $s$ matrices may be computed in any order. For each
ordering, consider the total cost of all matrix multiplications. Find
the minimal cost when the sizes of the matrices
are given. The running time of the algorithm
should be bounded by a
polynomial over the number of factors ($s$).}%
 \index{matrix multiplication, optimal order}

     Example. Matrices of size $2\times3$, $3\times4$,
$4\times5$ can be multiplied in two different ways. The cost is
either $2\cdot3\cdot4 + 2\cdot4\cdot5 = 64$  or
$3\cdot4\cdot5 + 2\cdot3\cdot5 = 90$.

    \Solution. Suppose the first matrix is associated with an
interval $[0,1]$, the second one is associated with
$[1,2]$,$\,\ldots$, and the $s$-th matrix is associated with $[s-1,s]$.
Adjacent matrices (for segments $[i-1,i]$ and $[i,i+1]$)
have a common dimension so we can multiply them. Let us
denote this common dimension by $d[i]$. Therefore, the initial
data of our problem is an array $d[0],\ldots,d[s]$.

    Let $a(i,j)$ be the minimal cost of computation of the
product of all the matrices in the interval $[i,j]$ (here $0\le
i<j\le s$). The cost in question is $a(0,s)$. The values of
$a(i,i+1)$ are equal to $0$ (we have only one matrix and nothing
to multiply). The recurrence formula is as follows:
 $$
    a(i,j) = \min\{a(i,k)+ a(k,j) + d[i]d[k]d[j]\}
 $$
where the minimal value is computed over all possible places of
the last multiplication, that is, over all $k=i+1,\ldots,j-1$.
Indeed, the product of all matrices in the interval $[i,k]$ is a
matrix of size $d[i]\times d[k]$, the product of all the
matrices in the interval $[k,j]$ has size $d[k]\times d[j]$, and
the cost of multiplication is $d[i]d[k]d[j]$.\nobreak\es

     Remark. The last two
problems are rather similar.
This is clear if we associate matrix factors with the sides
$1$--$2$, $2$--$3$,\ldots,$(s-1)$--$s$ of a polygon, and associate
any chord $i$--$j$ with the product of all matrices covered by
this chord.

\Problem{%
     A one-way railway has $n$ stops. We know the price of
tickets from the $i$-th stop to the $j$-th stop (for $i < j$, since
there is no traffic in the other direction). Find the minimal travel
cost from stop $1$ to stop $n$ (taking into account possible
savings due to intermediate stops).\es}

\medskip

     We have seen that sometimes we get a more effective
algorithm by replacing the recursion with a table that
is filled cell
by cell. A similar effect is achieved if we retain the recursive
algorithm, but store
the values of the
function already computed and do not compute them again when
the second request occurs. This trick is called
{\it memoization}.\index{memoization}

\Problem{%
     A finite set and a binary operation
$\langle u,v\rangle\mapsto u\circ v$
defined on this set are given (the
operation may be noncommutative and nonassociative). We have
$n$ elements $a_1,\ldots,a_n$ from the set and one more element
$x$. Check if it is possible to insert parentheses in the
expression $a_1\circ\cdots\circ a_n$ in such a way that the
result is equal to $x$. The number of operations should not
exceed $Cn^3$ for some constant $C$ (which depends on the
cardinality of the set given).}%
 \index{operation!non-associative}%
 \index{nonassociative operation}

     \Solution. Fill a table that contains (for any
subexpression
$a_i\circ\cdots\circ a_j$) the set of all
possible values
(for different placements of parentheses).\es

\smallskip

     The same trick is used in the polynomial algorithm that
tests whether a given word belongs to a context-free
language (see section~\ref{generalparsing},
p.~\pageref{generalparsingalgorithm}).

     The next problem (knapsack problem) was mentioned in
section~\ref{traversalusage}, p.~\pageref{traversalknapsack}.%
 \index{knapsack problem}%
 \index{problem!knapsack}%
 \label{dynamicknapsack}

\Problem{%
      An array
$x_1,\ldots,x_n$
of $n$ positive integers
and an
integer $N$ are given.
Check if $N$ is equal to the sum of some subset of
$\{x_1,\ldots,x_n\}$. The number of operations should be of
order $Nn$.}

    \Hint{After $i$ iterations, keep the set of all numbers
in $0,\ldots,N$ that can be represented as a sum of some
subset of
$\{x_1\ldots x_i\}$.}

