\subsection{Connected components, breadth and depth search}
\label{connectedcomponents}

     The simplest possible case of the shortest path problem is
when all the flight costs are $0$ or $+\infty$. This means that
we want to know whether it is possible to travel from $i$ to
$j$, but do not worry about the price. In other words, we have a
directed graph (a picture composed of points and arrows that
connect some of the points) and we want to know which points are
reachable from a given point via the arrows.%
 \index{graph!connected component}%
 \index{connected component!directed graph}

     For this special
case the algorithms given in the
preceding section are not optimal. Indeed, a faster recursive
program that solves this problem was given in
section~\ref{recursion}; its nonrecursive version was shown in
section~\ref{types}. Now we add the following
additional requirement: We
not only want
to list all the points (vertices) that are reachable
from a given vertex via arrows (edges), but we
also want to list them
in a specific order. Two of the
most popular instances of this
are the so-called
``breadth-first'' and ``depth-first'' search.

\bigskip

     {\bf Breadth-first search}%
 \index{search!breadth-first}

\medskip

We are to list all the vertices of a directed graph that are
reachable from a given vertex. The order is determined by the
distance (minimal number of edges between a vertex and the
given vertex).
This list solves the
answer to the problem of minimal travel cost when all the
edges have cost $1$ or $+\infty$.

\Problem{%
     Find an algorithm that performs breadth-first search in
time $Cm{}$, where $m$ is the total number of outgoing
edges of all reachable vertices.}

    \Solution.\label{breadthfirst}
This problem was considered in section~\ref{types},
p.~\pageref{breadth}. Here we give a detailed solution. Let
\w{num[i]} be the number of outgoing edges for vertex \w{i},
and let $\w{out[i][1]},\ldots,\w{out[i][num[i]]}$ be the terminal
vertices of the edges emanating from vertex \w{i}. Here is the
program (as it was written before):
\begin{verbatim}
  procedure Print_Reachable (i: integer);
  |   {print all the vertices reachable from i,
  |       including the vertex i itself}
  | var  X: subset of 1..n;
  |      P: subset of 1..n;
  |      q, v, w: 1..n;
  |      k: integer;
  begin
  | ...make X and P empty;
  | writeln (i);
  | ...add i to X, P;
  | {(1) P = is a set of printed vertices; P contains i;
  |  (2) only vertices reachable from i are printed;
  |  (3) X is a subset of P;
  |  (4) all printed vertices which have an outgoing edge to
  |      a non-printed vertex, belong to X}
  | while X is not empty do begin
  | | ...take some element of X into v;
  | | for k := 1 to num [v] do begin
  | | | w := out [v][k];
  | | | if w does not belong to P then begin
  | | | | writeln (w);
  | | | | add w to P;
  | | | | add w to X;
  | | | end;
  | | end;
  | end;
  end;
\end{verbatim}
If we don't worry about the order in which the reachable
vertices are printed, it doesn't matter which element of \w{X}
is chosen by the algorithm. Now we assume that \w{X} is a queue
(first in, first out). In this case, the program prints all
vertices reachable from \w{i} in order of increasing distance
from \w{i} (distance is the number of vertices on the shortest
path from \w{i}). Let us prove this assertion.

     By $V(k)$ we denote the set of vertices whose distance from
\w{i} (in the sense described above)
is $k$. The set $V(k+1)$ is equal to the set
$$
 \hbox{\rm (endpoints of edges whose startpoints are in
$V(k)$)}
\setminus (V(0) \cup \cdots \cup V(k))
$$
Let us prove now that for a nonnegative integer
$k=0,1,2\ldots$ there exists a
point during the execution of the program
(after one of the
\w{while}-iterations) such that
\begin{itemize}
\item the queue contains all the elements of $V(k)$ and no other
        elements;
\item all elements of $V(0),\ldots,V(k)$ and no others are printed.
\end{itemize}
For $k=0$, it is the state before the first iteration. Now comes
the
induction step: Assume that at some point,
the queue contains elements of $V(k)$.
Those elements are processed one by one (the new elements are
appended to the end of the queue and therefore cannot
interfere). The endpoints of the edges emanating from the
elements of $V(k)$ are printed and placed in the queue (unless they
were printed earlier), exactly as in the equation for
$V(k+1)$ shown above. Therefore, when all elements of $V(k)$ are
processed, the queue is filled with all the elements of
$V(k+1)$.\es

\bigskip

     {\bf Depth-first search}%
 \index{search!depth-first}

\medskip

     When thinking about depth-first search, it is
convenient to represent a given graph as the image of a tree. Let
us explain what we mean by this.
Suppose some vertex $x$ of a
directed graph is given. Assume that all vertices are reachable
(via edges) from $x$. We construct a tree that may be called
the ``universal covering tree'' of the graph. Its root is the point
$x$, and it has the same outgoing edges as in the graph. The
endpoints of those edges are sons of the root. Now consider any
son $y$ of $x$ and all its outgoing edges. Their endpoints are the sons
of $y$ in the tree. The
difference between the graph and the
tree is that different paths from $x$ to the same
vertex of the graph now lead to different vertices of the
tree. In other words, the vertex of the universal covering tree
is a path in the graph starting from $x$. Its sons are paths
that are one edge longer. Please note that the tree is
infinite if the graph has (reachable) directed cycles.

     There exists a natural mapping from the universal covering
tree to the graph. For any vertex $y$ in the graph, the number of
preimages is the number of paths from $x$ to $y$ in the graph.
Therefore, if we visit the tree vertices in some order, we at
the same time visit the vertices of the graph (but some graph
vertices may be visited many times).

     Assume that for any graph vertex the outgoing edges are
numbered. Then for any vertex of the universal covering tree its
sons are numbered. Let us visit tree vertices in the following
order: first the root, then the subtrees rooted at the root's sons
(in the given order of sons). An algorithm which traverses
tree in that order was considered in section~\ref{recursion}. This
algorithm
can be modified to traverse the graph avoiding visits to
vertices already visited. Doing that, we get what is called
``depth-first search''.

Here is another description of
depth-first search.
Let us
introduce a linear ordering
on paths starting at a given
vertex $x$. Any path precedes all its extensions. If two paths
diverge at some vertex, they are ordered according to the
ordering of the outgoing edges at that vertex.
After that,
vertices are
ordered according to the minimal paths reaching them. This ordering
is called {\it depth-first\/} ordering.

\Problem{%
     Write an algorithm for depth-first search.}

    \Hint{Take a program that traverses a tree (root
$\rightarrow$ left subtree $\rightarrow$ right subtree) from
section~\ref{recursion} or~\ref{delrec} and modify it.
The main difference is that we do not want to revisit any
visited vertex. Therefore, if we are at an already-visited
vertex, we do nothing. (If a path is not minimal among
all paths going to some vertex, all its extensions are not minimal
as well, and can be safely ignored.)}

     Remark. Recall that in section~\ref{delrec} two possible
nonrecursive algorithms for tree traversal were mentioned
(p.~\pageref{twotraversals}). Both versions may be used for
depth-first search.

\medskip

    Depth-first search is used in several graph algorithms
(sometimes in a modified form).


\Problem{%
      An undirected graph is called a {\it bipartite graph\/} if its
vertices may be colored in two colors in such a way that
each edge connects vertices of different colors. Find an algorithm that
checks whether a graph is a bipartite graph in time
$C\cdot{}$(number of edges${}+{}$number of vertices).}%
 \index{graph!bipartite}%
 \index{bipartite graph}

    \Hint{(a)~Each connected component may be considered separately.
(b)~After we choose the color of some vertex, the colors of all other
vertices of the same component are uniquely
determined.}

     Remark. In this problem we may use breadth-first as
well as depth-first search.

\Problem{%
     Write a nonrecursive algorithm for topological sorting of a
directed graph without cycles. (For a recursive algorithm, see
p.~\pageref{rectopsort}.)}%
 \index{topological sorting}%
 \index{sorting!topological}

    \Solution. Assume that the graph has vertices
$\w{1..n}$. For any vertex \w{i}, we know the number
\w{num[i]} of outgoing edges and the vertices
$\w{dest[i][1]},\ldots,\w{dest[i][num[i]]}$ that the outgoing edges
enter. We adopt the following terminology: the outgoing edges are
listed ``from left to right'' (so \w{dest[i][1]} is ``on the left''
of \w{dest[i][2]}, etc.).

Our goal is to print all the vertices of the graph; the
requirement is that the endpoint of any edge is printed before
its starting point. We assume that the graph has no cycles
(otherwise this is impossible).

Let us add to the graph an auxiliary vertex \w{0} that has
\w{n} outgoing edges to $\w{1},\ldots,\w{n}$. If it is printed
and the requirement is fulfilled, then all other vertices are
already
printed.

Our algorithm maintains a path that starts at \w{0} (the auxiliary
vertex) and traverses the graph edges. The length of this path
is kept in an integer variable \w{m}. The path is formed by the
vertices $\w{vert[1]..vert[m]}$ and edges having numbers
$\w{edge[1]..edge[m]}$. The number \w{edge[s]} refers to
the numbering of all outgoing edges of the vertex $\w{vert[s]}$.
Therefore, for all \w{s}, the following inequality holds:
$$
     \w{edge[s]}\le\w{num[vert[s]]}
$$
as well as the equality
$$
     \w{vert[s+1]} = \w{dest}\,\w{[vert[s]]}\,\w{[edge[s]]}.
$$
Please note that the end of the last edge in the path (i.e.,
the vertex \w{dest[vert[m]][edge[m]]}), is not included in the
array \w{vert}. Moreover, we make an exception for the last edge
and allow it to point ``nowhere'': \w{edge[m]} may be equal
to \w{num[vert[m]]+1}.

The algorithm prints the vertices of the graph; a vertex is printed
only after all the vertices where the outgoing edges go are printed.
Moreover, the following requirement (I) is fulfilled:
\begin{quote}
     all vertices in the path, except the last one (i.e., the
     vertices $\w{vert[1]..vert[m]}$), are not printed,
     but if we turn to the left and leave our path, we immediately
     come to an already printed vertex.
\end{quote}
Here is the algorithm in full:
\begin{Verbatim}
        m:=1; vert[1]:=0; edge[1]:=1;
        {(I)}
        while not( (m=1) and (edge[1]=n+1)) do begin
        | if edge[m]=num[vert[m]]+1 then begin
        | | {path leads to nowhere, therefore all vertices
        | |     following vert[m] are printed and we may
        | |     print vert[m]}
        | | writeln (vert[m]);
        | | m:=m-1; edge[m]:=edge[m]+1;
        | end else begin
        | |  {edge[m] <= num[vert[m]], path ends in a real
        | |     vertex}
        | |  lastvert:= dest[vert[m]][edge[m]];
        | |  if lastvert is printed then begin
        | |  | edge[m]:=edge[m]+1;
        | |  end else begin
        | |  | m:=m+1; vert[m]:=lastvert; edge[m]:=1;
        | |  end;
        | end;
        end;
        {the path immediately goes to nowhere, so all the
         vertices on the left (1..n) are printed}
\end{Verbatim}

\Problem{%
     Prove that if the graph has no cycles, this algorithm terminates.}

     \Solution. Assume that this is not true. Any vertex may be
printed at most once, so the vertices are not printed after some
point. In a graph without cycles, the path length is limited
(no vertex can appear in a path twice); therefore, after some
point the path never becomes
longer. After that, the only possibility is an increase in
\w{edge[m]}, but this cannot happen infinitely many times.\es

\Problem{
   Prove that the running time of the previous
   algorithm is
   $O($number of vertices $+$ number of edges$)$.\es
}

\Problem{
   Modify the algorithm in such a way that it can be applied to
   any graph.  The algorithm should
   either find a cycle (if it exists) or perform a
   topological sort (if there are no cycles).\es
}


